import os
from typing import Any, Dict, Generator, List

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.base import BaseClient
from clarifai.client.deployment import Deployment
from clarifai.client.lister import Lister
from clarifai.errors import UserError
from clarifai.utils.logging import logger


class Nodepool(Lister, BaseClient):
  """Nodepool is a class that provides access to Clarifai API endpoints related to Nodepool information."""

  def __init__(self,
               nodepool_id: str = None,
               user_id: str = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes a Nodepool object.

    Args:
        nodepool_id (str): The Nodepool ID for the Nodepool to interact with.
        user_id (str): The user ID of the user.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the nodepool.
    """
    self.kwargs = {**kwargs, 'id': nodepool_id}
    self.nodepool_info = resources_pb2.Nodepool(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=user_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def list_deployments(self,
                       filter_by: Dict[str, Any] = {},
                       page_no: int = None,
                       per_page: int = None) -> Generator[Deployment, None, None]:
    """Lists all the available deployments of compute cluster.

    Args:
        filter_by (Dict[str, Any]): The filter to apply to the list of deployments.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Deployment: Deployment objects for the nodepools in the compute cluster.

    Example:
        >>> from clarifai.client.nodepool import Nodepool
        >>> nodepool = Nodepool(nodepool_id="nodepool_id", user_id="user_id")
        >>> all_deployments = list(nodepool.list_deployments())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(user_app_id=self.user_app_id, nodepool_id=self.id, **filter_by)
    all_deployments_info = self.list_pages_generator(
        self.STUB.ListDeployments,
        service_pb2.ListDeploymentsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for deployment_info in all_deployments_info:
      yield Deployment.from_auth_helper(auth=self.auth_helper, **deployment_info)

  def _process_deployment_config(self, config_filepath: str) -> Dict[str, Any]:
    with open(config_filepath, "r") as file:
      deployment_config = yaml.safe_load(file)

    assert "deployment" in deployment_config, "deployment info not found in the config file"
    deployment = deployment_config['deployment']
    assert "autoscale_config" in deployment, "autoscale_config not found in the config file"
    assert ("worker" in deployment) and (
        ("model" in deployment["worker"]) or
        ("workflow" in deployment["worker"])), "worker info not found in the config file"
    assert "scheduling_choice" in deployment, "scheduling_choice not found in the config file"
    assert "nodepools" in deployment, "nodepools not found in the config file"
    deployment['user_id'] = self.user_app_id.user_id
    deployment['autoscale_config'] = resources_pb2.AutoscaleConfig(
        **deployment['autoscale_config'])
    deployment['nodepools'] = [
        resources_pb2.Nodepool(
            id=nodepool['id'],
            compute_cluster=resources_pb2.ComputeCluster(
                id=nodepool['compute_cluster']['id'], user_id=self.user_app_id.user_id))
        for nodepool in deployment['nodepools']
    ]
    if 'user' in deployment['worker']:
      deployment['worker']['user'] = resources_pb2.User(**deployment['worker']['user'])
    elif 'model' in deployment['worker']:
      deployment['worker']['model'] = resources_pb2.Model(**deployment['worker']['model'])
    elif 'workflow' in deployment['worker']:
      deployment['worker']['workflow'] = resources_pb2.Workflow(**deployment['worker']['workflow'])
    deployment['worker'] = resources_pb2.Worker(**deployment['worker'])
    if "visibility" in deployment:
      deployment["visibility"] = resources_pb2.Visibility(**deployment["visibility"])
    return deployment

  @staticmethod
  def get_runner_selector(user_id: str, compute_cluster_id: str,
                          nodepool_id: str) -> resources_pb2.RunnerSelector:
    """Returns a RunnerSelector object for the specified compute cluster and nodepool.

    Args:
        user_id (str): The user ID of the user.
        compute_cluster_id (str): The compute cluster ID for the compute cluster.
        nodepool_id (str): The nodepool ID for the nodepool.

    Returns:
        resources_pb2.RunnerSelector: A RunnerSelector object for the specified compute cluster and nodepool.

    Example:
        >>> from clarifai.client.nodepool import Nodepool
        >>> nodepool = Nodepool(nodepool_id="nodepool_id", user_id="user_id")
        >>> runner_selector = Nodepool.get_runner_selector(user_id="user_id", compute_cluster_id="compute_cluster_id", nodepool_id="nodepool_id")
    """
    compute_cluster = resources_pb2.ComputeCluster(id=compute_cluster_id, user_id=user_id)
    nodepool = resources_pb2.Nodepool(id=nodepool_id, compute_cluster=compute_cluster)
    return resources_pb2.RunnerSelector(nodepool=nodepool)

  def create_deployment(self, config_filepath: str, deployment_id: str = None) -> Deployment:
    """Creates a deployment for the nodepool.

    Args:
        config_filepath (str): The path to the deployment config file.
        deployment_id (str): New deployment ID for the deployment to create.

    Returns:
        Deployment: A Deployment object for the specified deployment ID.

    Example:
        >>> from clarifai.client.nodepool import Nodepool
        >>> nodepool = Nodepool(nodepool_id="nodepool_id", user_id="user_id")
        >>> deployment = nodepool.create_deployment(config_filepath="config.yml")
    """
    if not os.path.exists(config_filepath):
      raise UserError(f"Deployment config file not found at {config_filepath}")

    deployment_config = self._process_deployment_config(config_filepath)

    if 'id' in deployment_config:
      if deployment_id is None:
        deployment_id = deployment_config['id']
      deployment_config.pop('id')

    request = service_pb2.PostDeploymentsRequest(
        user_app_id=self.user_app_id,
        deployments=[resources_pb2.Deployment(id=deployment_id, **deployment_config)])
    response = self._grpc_request(self.STUB.PostDeployments, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nDeployment created\n%s", response.status)

    return Deployment.from_auth_helper(self.auth_helper, deployment_id=deployment_id)

  def deployment(self, deployment_id: str) -> Deployment:
    """Returns a Deployment object for the existing deployment ID.

    Args:
        deployment_id (str): The deployment ID for the deployment to interact with.

    Returns:
        Deployment: A Deployment object for the existing deployment ID.

    Example:
        >>> from clarifai.client.nodepool import Nodepool
        >>> nodepool = Nodepool(nodepool_id="nodepool_id", user_id="user_id")
        >>> deployment = nodepool.deployment(deployment_id="deployment_id")
    """
    request = service_pb2.GetDeploymentRequest(
        user_app_id=self.user_app_id, deployment_id=deployment_id)
    response = self._grpc_request(self.STUB.GetDeployment, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response[list(dict_response.keys())[1]],
                                        list(dict_response.keys())[1])

    return Deployment.from_auth_helper(auth=self.auth_helper, **kwargs)

  def delete_deployments(self, deployment_ids: List[str]) -> None:
    """Deletes list of deployments for the nodepool.

    Args:
        deployment_ids (List[str]): The list of deployment IDs to delete.

    Example:
        >>> from clarifai.client.nodepool import Nodepool
        >>> nodepool = Nodepool(nodepool_id="nodepool_id", user_id="user_id")
        >>> nodepool.delete_deployments(deployment_ids=["deployment_id1", "deployment_id2"])
    """
    assert isinstance(deployment_ids, list), "deployment_ids param should be a list"

    request = service_pb2.DeleteDeploymentsRequest(
        user_app_id=self.user_app_id, ids=deployment_ids)
    response = self._grpc_request(self.STUB.DeleteDeployments, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nDeployments Deleted\n%s", response.status)

  def __getattr__(self, name):
    return getattr(self.nodepool_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.nodepool_info, param)}" for param in init_params
        if hasattr(self.nodepool_info, param)
    ]
    return f"Nodepool Details: \n{', '.join(attribute_strings)}\n"



import os

registry = os.environ.get('CLARIFAI_BASE_IMAGE_REGISTRY', 'public.ecr.aws/clarifai-models')

GIT_SHA = "b8ae56bf3b7c95e686ca002b07ca83d259c716eb"

PYTHON_BASE_IMAGE = registry + '/python-base:{python_version}-' + GIT_SHA
TORCH_BASE_IMAGE = registry + '/torch:{torch_version}-py{python_version}-{gpu_version}-' + GIT_SHA

# List of available python base images
AVAILABLE_PYTHON_IMAGES = ['3.11', '3.12']

DEFAULT_PYTHON_VERSION = 3.12

# By default we download at runtime.
DEFAULT_DOWNLOAD_CHECKPOINT_WHEN = "runtime"

# Folder for downloading checkpoints at runtime.
DEFAULT_RUNTIME_DOWNLOAD_PATH = os.path.join(os.sep, "tmp", ".cache")

# List of available torch images
# Keep sorted by most recent cuda version.
AVAILABLE_TORCH_IMAGES = [
    '2.4.1-py3.11-cu124',
    '2.5.1-py3.11-cu124',
    '2.4.1-py3.12-cu124',
    '2.5.1-py3.12-cu124',
    '2.6.0-py3.12-cu126',
    '2.7.0-py3.12-cu128',
    '2.7.0-py3.12-rocm6.3',
]
CONCEPTS_REQUIRED_MODEL_TYPE = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'text-classifier'
]



import fnmatch
import importlib.util
import json
import os
import shutil

import requests

from clarifai.runners.utils.const import CONCEPTS_REQUIRED_MODEL_TYPE
from clarifai.utils.logging import logger


class HuggingFaceLoader:

  HF_DOWNLOAD_TEXT = "The 'huggingface_hub' package is not installed. Please install it using 'pip install huggingface_hub'."

  def __init__(self, repo_id=None, token=None, model_type_id=None):
    self.repo_id = repo_id
    self.token = token
    self.clarifai_model_type_id = model_type_id
    if token:
      if self.validate_hftoken(token):
        try:
          from huggingface_hub import login
        except ImportError:
          raise ImportError(self.HF_DOWNLOAD_TEXT)
        login(token=token)
        logger.info("Hugging Face token validated")
      else:
        self.token = None
        logger.info("Continuing without Hugging Face token")

  @classmethod
  def validate_hftoken(cls, hf_token: str):
    try:
      if importlib.util.find_spec("huggingface_hub") is None:
        raise ImportError(cls.HF_DOWNLOAD_TEXT)
      from huggingface_hub import HfApi

      api = HfApi()
      api.whoami(token=hf_token)
      return True
    except Exception as e:
      logger.error(
          f"Error setting up Hugging Face token, please make sure you have the correct token: {e}")
      return False

  def download_checkpoints(self,
                           checkpoint_path: str,
                           allowed_file_patterns=None,
                           ignore_file_patterns=None):
    # throw error if huggingface_hub wasn't installed
    try:
      from huggingface_hub import snapshot_download
    except ImportError:
      raise ImportError(self.HF_DOWNLOAD_TEXT)
    if os.path.exists(checkpoint_path) and self.validate_download(
        checkpoint_path, allowed_file_patterns, ignore_file_patterns):
      logger.info("Checkpoints already exist")
      return True
    else:
      os.makedirs(checkpoint_path, exist_ok=True)
      try:
        is_hf_model_exists = self.validate_hf_model()
        if not is_hf_model_exists:
          logger.error("Model %s not found on Hugging Face" % (self.repo_id))
          return False

        self.ignore_patterns = self._get_ignore_patterns()
        if ignore_file_patterns:
          if self.ignore_patterns:
            self.ignore_patterns.extend(ignore_file_patterns)
          else:
            self.ignore_patterns = ignore_file_patterns
        snapshot_download(
            repo_id=self.repo_id,
            local_dir=checkpoint_path,
            local_dir_use_symlinks=False,
            allow_patterns=allowed_file_patterns,
            ignore_patterns=self.ignore_patterns)
        # Remove the `.cache` folder if it exists
        cache_path = os.path.join(checkpoint_path, ".cache")
        if os.path.exists(cache_path) and os.path.isdir(cache_path):
          shutil.rmtree(cache_path)

      except Exception as e:
        logger.error(f"Error downloading model checkpoints {e}")
        return False
      finally:
        is_downloaded = self.validate_download(checkpoint_path, allowed_file_patterns,
                                               ignore_file_patterns)
        if not is_downloaded:
          logger.error("Error validating downloaded model checkpoints")
          return False
      return True

  def download_config(self, checkpoint_path: str):
    # throw error if huggingface_hub wasn't installed
    try:
      from huggingface_hub import hf_hub_download
    except ImportError:
      raise ImportError(self.HF_DOWNLOAD_TEXT)
    if os.path.exists(checkpoint_path) and os.path.exists(
        os.path.join(checkpoint_path, 'config.json')):
      logger.info("HF model's config.json already exists")
      return True
    os.makedirs(checkpoint_path, exist_ok=True)
    try:
      is_hf_model_exists = self.validate_hf_model()
      if not is_hf_model_exists:
        logger.error("Model %s not found on Hugging Face" % (self.repo_id))
        return False
      hf_hub_download(repo_id=self.repo_id, filename='config.json', local_dir=checkpoint_path)
    except Exception as e:
      logger.error(f"Error downloading model's config.json {e}")
      return False
    return True

  def validate_hf_model(self,):
    # check if model exists on HF
    try:
      from huggingface_hub import file_exists, repo_exists
    except ImportError:
      raise ImportError(self.HF_DOWNLOAD_TEXT)
    if self.clarifai_model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE:
      return repo_exists(self.repo_id) and file_exists(self.repo_id, 'config.json')
    else:
      return repo_exists(self.repo_id)

  def validate_download(self, checkpoint_path: str, allowed_file_patterns: list,
                        ignore_file_patterns: list):
    # check if model exists on HF
    try:
      from huggingface_hub import list_repo_files
    except ImportError:
      raise ImportError(self.HF_DOWNLOAD_TEXT)
    # Get the list of files on the repo
    repo_files = list_repo_files(self.repo_id, token=self.token)

    # Get the list of files on the repo that are allowed
    if allowed_file_patterns:

      def should_allow(file_path):
        return any(fnmatch.fnmatch(file_path, pattern) for pattern in allowed_file_patterns)

      repo_files = [f for f in repo_files if should_allow(f)]

    self.ignore_patterns = self._get_ignore_patterns()
    if ignore_file_patterns:
      if self.ignore_patterns:
        self.ignore_patterns.extend(ignore_file_patterns)
      else:
        self.ignore_patterns = ignore_file_patterns
    # Get the list of files on the repo that are not ignored
    if getattr(self, "ignore_patterns", None):
      patterns = self.ignore_patterns

      def should_ignore(file_path):
        return any(fnmatch.fnmatch(file_path, pattern) for pattern in patterns)

      repo_files = [f for f in repo_files if not should_ignore(f)]

    # Check if downloaded files match the files we expect (ignoring ignored patterns)
    checkpoint_dir_files = []
    for dp, dn, fn in os.walk(os.path.expanduser(checkpoint_path)):
      checkpoint_dir_files.extend(
          [os.path.relpath(os.path.join(dp, f), checkpoint_path) for f in fn])

    # Validate by comparing file lists
    return len(checkpoint_dir_files) >= len(repo_files) and not (
        len(set(repo_files) - set(checkpoint_dir_files)) > 0) and len(repo_files) > 0

  def _get_ignore_patterns(self):
    # check if model exists on HF
    try:
      from huggingface_hub import list_repo_files
    except ImportError:
      raise ImportError(self.HF_DOWNLOAD_TEXT)

    # Get the list of files on the repo that are not ignored
    repo_files = list_repo_files(self.repo_id, token=self.token)
    self.ignore_patterns = None
    if any(f.endswith(".safetensors") for f in repo_files):
      self.ignore_patterns = [
          "**/original/*", "**/*.pth", "**/*.bin", "*.pth", "*.bin", "**/.cache/*"
      ]
    return self.ignore_patterns

  @staticmethod
  def validate_config(checkpoint_path: str):
    # check if downloaded config.json exists
    return os.path.exists(checkpoint_path) and os.path.exists(
        os.path.join(checkpoint_path, 'config.json'))

  @staticmethod
  def validate_concept(checkpoint_path: str):
    # check if downloaded concept exists in hf model
    config_path = os.path.join(checkpoint_path, 'config.json')
    with open(config_path, 'r') as f:
      config = json.load(f)

    labels = config.get('id2label', None)
    if labels:
      return True
    return False

  @staticmethod
  def fetch_labels(checkpoint_path: str):
    # Fetch labels for classification, detection and segmentation models
    config_path = os.path.join(checkpoint_path, 'config.json')
    with open(config_path, 'r') as f:
      config = json.load(f)

    labels = config['id2label']
    return labels

  @staticmethod
  def get_huggingface_checkpoint_total_size(repo_name):
    """
    Fetches the JSON data for a Hugging Face model using the API with `?blobs=true`.
    Calculates the total size from the JSON output.

    Args:
        repo_name (str): The name of the model on Hugging Face Hub. e.g. "casperhansen/llama-3-8b-instruct-awq"

    Returns:
        int: The total size in bytes.
    """
    try:
      url = f"https://huggingface.co/api/models/{repo_name}?blobs=true"
      response = requests.get(url)
      response.raise_for_status()  # Raise an exception for bad status codes
      json_data = response.json()

      if isinstance(json_data, str):
        data = json.loads(json_data)
      else:
        data = json_data

      total_size = 0
      for file in data['siblings']:
        total_size += file['size']
      return total_size
    except Exception as e:
      logger.error(f"Error fetching checkpoint size from huggingface.co: {e}")
      return 0



import os

CLARIFAI_PAT_ENV_VAR = "CLARIFAI_PAT"
CLARIFAI_SESSION_TOKEN_ENV_VAR = "CLARIFAI_SESSION_TOKEN"
CLARIFAI_USER_ID_ENV_VAR = "CLARIFAI_USER_ID"

DEFAULT_CONFIG = f'{os.environ["HOME"]}/.config/clarifai/config'



import click

from clarifai.cli.base import cli
from clarifai.client.user import User
from clarifai.utils.cli import AliasedGroup, display_co_resources, validate_context


@cli.group(['computecluster', 'cc'], cls=AliasedGroup)
def computecluster():
  """Manage Compute Clusters: create, delete, list"""


@computecluster.command(['c'])
@click.argument('compute_cluster_id')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the compute cluster config file.')
@click.pass_context
def create(ctx, compute_cluster_id, config):
  """Create a new Compute Cluster with the given config file."""
  validate_context(ctx)
  user = User(
      user_id=ctx.obj.current.user_id, pat=ctx.obj.current.pat, base_url=ctx.obj.current.api_base)
  if compute_cluster_id:
    user.create_compute_cluster(config, compute_cluster_id=compute_cluster_id)
  else:
    user.create_compute_cluster(config)


@computecluster.command(['ls'])
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, page_no, per_page):
  """List all compute clusters for the user."""
  validate_context(ctx)
  user = User(
      user_id=ctx.obj.current.user_id, pat=ctx.obj.current.pat, base_url=ctx.obj.current.api_base)
  response = user.list_compute_clusters(page_no, per_page)
  display_co_resources(
      response,
      custom_columns={
          'ID': lambda c: c.id,
          'USER_ID': lambda c: c.user_id,
          'DESCRIPTION': lambda c: c.description,
      })


@computecluster.command(['rm'])
@click.argument('compute_cluster_id')
@click.pass_context
def delete(ctx, compute_cluster_id):
  """Deletes a compute cluster for the user."""
  validate_context(ctx)
  user = User(
      user_id=ctx.obj.current.user_id, pat=ctx.obj.current.pat, base_url=ctx.obj.current.api_base)
  user.delete_compute_clusters([compute_cluster_id])



from unittest import mock

import grpc
import pytest as pytest
from clarifai_grpc.grpc.api import service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.auth.helper import ClarifaiAuthHelper, clear_cache
from clarifai.client.auth.stub import AuthorizedStub, RetryStub


class MockRpcError(grpc.RpcError):
  pass


@pytest.fixture(autouse=True)
def clear_caches():
  clear_cache()


def test_auth_unary_unary():
  auth = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  stub = AuthorizedStub(auth)
  with mock.patch.object(stub.stub, 'ListInputs', spec=stub.stub.ListInputs) as mock_f:
    req = service_pb2.ListInputsRequest()
    req.user_app_id.app_id = 'test_auth_unary_unary'
    stub.ListInputs(req)
    mock_f.assert_called_with(req, metadata=auth.metadata)


def test_auth_unary_unary_future():
  auth = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  stub = AuthorizedStub(auth)
  with mock.patch.object(stub.stub, 'ListInputs', spec=stub.stub.ListInputs) as mock_f:
    req = service_pb2.ListInputsRequest()
    req.user_app_id.app_id = 'test_auth_unary_unary_future'
    stub.ListInputs.future(req)
    mock_f.future.assert_called_with(req, metadata=auth.metadata)


def test_auth_unary_stream():
  auth = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  stub = AuthorizedStub(auth)
  with mock.patch.object(stub.stub, 'StreamInputs', spec=stub.stub.StreamInputs) as mock_f:
    req = service_pb2.StreamInputsRequest()
    req.user_app_id.app_id = 'test_auth_unary_stream'
    stub.StreamInputs(req)
    mock_f.assert_called_with(req, metadata=auth.metadata)


def test_retry_unary_unary():
  max_attempts = 5
  auth = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  stub = RetryStub(AuthorizedStub(auth), max_attempts=max_attempts, backoff_time=0.0001)
  retry_response = service_pb2.MultiInputResponse()
  retry_response.status.code = status_code_pb2.CONN_THROTTLED
  success_response = service_pb2.MultiInputResponse()
  success_response.status.code = status_code_pb2.SUCCESS
  for nfailures in range(0, max_attempts + 1):
    mock_resps = [retry_response] * nfailures + [success_response]
    with mock.patch.object(
        stub.stub, 'ListInputs', spec=stub.stub.stub.ListInputs, side_effect=mock_resps) as mock_f:
      req = service_pb2.ListInputsRequest()
      req.user_app_id.app_id = 'test_retry_unary_unary'
      res = stub.ListInputs(req)
      assert mock_f.call_count == min(max_attempts, len(mock_resps))
      if nfailures < max_attempts:
        assert res is success_response
      else:
        assert res is retry_response


def test_retry_grpcconn_unary_unary():
  max_attempts = 5
  auth = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  stub = RetryStub(AuthorizedStub(auth), max_attempts=max_attempts, backoff_time=0.0001)
  retry_response = service_pb2.MultiInputResponse()
  retry_response.status.code = status_code_pb2.CONN_THROTTLED
  success_response = service_pb2.MultiInputResponse()
  success_response.status.code = status_code_pb2.SUCCESS
  error = MockRpcError()
  error.code = lambda: grpc.StatusCode.UNAVAILABLE
  for nfailures in range(0, max_attempts + 1):
    mock_resps = [error] * nfailures + [success_response]
    with mock.patch.object(
        stub.stub, 'ListInputs', spec=stub.stub.stub.ListInputs, side_effect=mock_resps) as mock_f:
      req = service_pb2.ListInputsRequest()
      req.user_app_id.app_id = 'test_retry_unary_unary'
      try:
        res = stub.ListInputs(req)
      except Exception as e:
        res = e
      assert mock_f.call_count == min(max_attempts, len(mock_resps))
      if nfailures < max_attempts:
        assert res is success_response
      else:
        assert res is error



from typing import Dict, List, Tuple, Union

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from PIL import Image

from clarifai.client.auth.helper import ClarifaiAuthHelper

from .data_utils import bytes_to_image, image_to_bytes


class BaseDataHandler:

  def __init__(self,
               proto: Union[resources_pb2.Input, resources_pb2.Output],
               auth: ClarifaiAuthHelper = None):
    self._proto = proto
    self._auth = auth

  #
  def to_python(self):
    return dict(text=self.text, image=self.image, audio=self.audio)

  # ---------------- Start get/setters ---------------- #
  # Proto
  @property
  def proto(self):
    return self._proto

  # Status
  @property
  def status(self) -> status_pb2.Status:
    return self._proto.status

  def set_status(self, code: str, description: str = ""):
    self._proto.status.code = code
    self._proto.status.description = description

  # Text
  @property
  def text(self) -> Union[None, str]:
    data = self._proto.data.text
    text = None
    if data.ByteSize():
      if data.raw:
        text = data.raw
      else:
        raise NotImplementedError
    return text

  def set_text(self, text: str):
    self._proto.data.text.raw = text

  # Image
  @property
  def image(self, format: str = "np") -> Union[None, Image.Image, np.ndarray]:
    data = self._proto.data.image
    image = None
    if data.ByteSize():
      data: resources_pb2.Image = data
      if data.base64:
        image = data.base64
      elif data.url:
        raise NotImplementedError
      image = bytes_to_image(image)
      image = image if not format == "np" else np.asarray(image).astype("uint8")
    return image

  def set_image(self, image: Union[Image.Image, np.ndarray]):
    if isinstance(image, np.ndarray):
      image = Image.fromarray(image)
    self._proto.data.image.base64 = image_to_bytes(image)

  # Audio
  @property
  def audio(self) -> bytes:
    data = self._proto.data.audio
    audio = None
    if data.ByteSize():
      if data.base64:
        audio = data.base64
    return audio

  def set_audio(self, audio: bytes):
    self._proto.data.audio.base64 = audio

  # Bboxes
  @property
  def bboxes(self, real_coord: bool = False, image_width: int = None,
             image_height: int = None) -> Tuple[List, List, List]:
    if real_coord:
      assert (image_height or image_width
             ), "image_height and image_width are required when when return real coordinates"
    xyxy = []
    scores = []
    concepts = []
    for _, each in enumerate(self._proto.data.regions):
      box = each.region_info
      score = each.value
      concept = each.data.concepts[0].id
      x1 = box.left_col
      y1 = box.top_row
      x2 = box.right_col
      y2 = box.bottom_row
      if real_coord:
        x1 = x1 * image_width
        y1 = y1 * image_height
        x2 = x2 * image_width
        y2 = y2 * image_height
      xyxy.append([x1, y1, x2, y2])
      scores.append(score)
      concepts.append(concept)

    return xyxy, scores, concepts

  def set_bboxes(self,
                 boxes: list,
                 scores: list,
                 concepts: list,
                 real_coord: bool = False,
                 image_width: int = None,
                 image_height: int = None):
    if real_coord:
      assert (image_height and
              image_width), "image_height and image_width are required when `real_coord` is set"
      bboxes = [[x[1] / image_height, x[0] / image_width, x[3] / image_height, x[2] / image_width]
                for x in boxes]  # normalize the bboxes to [0,1] and [y1 x1 y2 x2]
      bboxes = np.clip(bboxes, 0, 1.0)

    regions = []
    for ith, bbox in enumerate(bboxes):
      score = scores[ith]
      concept = concepts[ith]
      if any([each > 1.0 for each in bbox]):
        assert ValueError(
            "Box coordinates is not normalized between [0, 1]. Please set format_box to True and provide image_height and image_width to normalize"
        )
      region = resources_pb2.RegionInfo(bounding_box=resources_pb2.BoundingBox(
          top_row=bbox[0],  # y_min
          left_col=bbox[1],  # x_min
          bottom_row=bbox[2],  # y_max
          right_col=bbox[3],  # x_max
      ))
      data = resources_pb2.Data(concepts=resources_pb2.Concept(id=concept, value=score))
      regions.append(resources_pb2.Region(region_info=region, data=data))

    self._proto.data.regions = regions

  # Concepts
  @property
  def concepts(self) -> Dict[str, float]:
    con_scores = {}
    for each in self.proto.data.concepts:
      con_scores.update({each.id: each.value})
    return con_scores

  def set_concepts(self, concept_score_pairs: Dict[str, float]):
    concepts = []
    for concept, score in concept_score_pairs.items():
      con_score = resources_pb2.Concept(id=concept, name=concept, value=score)
      concepts.append(con_score)
    if concepts:
      self._proto.data.ClearField("concepts")
      for each in concepts:
        self._proto.data.concepts.append(each)

  # Embeddings
  @property
  def embeddings(self) -> List[List[float]]:
    return [each.vector for each in self.proto.data.embeddings]

  def set_embeddings(self, list_vectors: List[List[float]]):
    if list_vectors[0]:
      self._proto.data.ClearField("embeddings")
    for vec in list_vectors:
      self._proto.data.embeddings.append(
          resources_pb2.Embedding(vector=vec, num_dimensions=len(vec)))

  # ---------------- End get/setters ---------------- #

  # Constructors
  @classmethod
  def from_proto(cls, proto):
    clss = cls(proto=proto)
    return clss

  @classmethod
  def from_data(
      cls,
      status_code: int = status_code_pb2.SUCCESS,
      status_description: str = "",
      text: str = None,
      image: Union[Image.Image, np.ndarray] = None,
      audio: bytes = None,
      boxes: dict = None,
      concepts: Dict[str, float] = {},
      embeddings: List[List[float]] = [],
  ) -> 'OutputDataHandler':
    clss = cls(proto=resources_pb2.Output())
    if isinstance(image, Image.Image) or isinstance(image, np.ndarray):
      clss.set_image(image)
    if text:
      clss.set_text(text)
    if audio:
      clss.set_audio(audio)
    if boxes:
      clss.set_bboxes(**boxes)
    if concepts:
      clss.set_concepts(concepts)
    if embeddings:
      clss.set_embeddings(embeddings)

    clss.set_status(code=status_code, description=status_description)
    return clss


class InputDataHandler(BaseDataHandler):

  def __init__(self,
               proto: resources_pb2.Input = resources_pb2.Input(),
               auth: ClarifaiAuthHelper = None):
    super().__init__(proto=proto, auth=auth)


class OutputDataHandler(BaseDataHandler):

  def __init__(self,
               proto: resources_pb2.Output = resources_pb2.Output(),
               auth: ClarifaiAuthHelper = None):
    super().__init__(proto=proto, auth=auth)



import os
from collections import OrderedDict
from dataclasses import dataclass, field

import yaml

from clarifai.utils.constants import DEFAULT_CONFIG


class Context(OrderedDict):
  """
    A context which has a name and a set of key-values as a dict under env.

    You can access the keys directly.
    """

  def __init__(self, name, **kwargs):
    self['name'] = name
    # when loading from config we may have the env: section in yaml already so we get it here.
    if 'env' in kwargs:
      self['env'] = kwargs['env']
    else:  # when consructing as Context(name, key=value) we set it here.
      self['env'] = kwargs

  def __getattr__(self, key):
    try:
      if key == 'name':
        return self[key]
      if key == 'env':
        raise AttributeError("Don't access .env directly")

      # Allow accessing CLARIFAI_PAT type env var names from config as .pat
      envvar_name = 'CLARIFAI_' + key.upper()
      env = self['env']
      if envvar_name in env:
        value = env[envvar_name]
        if value == "ENVVAR":
          return os.environ[envvar_name]
      else:
        value = env[key]

      if isinstance(value, dict):
        return Context(value)

      return value
    except KeyError as e:
      raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'") from e

  def __setattr__(self, key, value):
    if key == "name":
      self['name'] = value
    else:
      self['env'][key] = value

  def __delattr__(self, key):
    try:
      del self['env'][key]
    except KeyError as e:
      raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'") from e

  def to_serializable_dict(self):
    return dict(self['env'])


@dataclass
class Config():
  current_context: str
  filename: str
  contexts: OrderedDict[str, Context] = field(default_factory=OrderedDict)

  def __post_init__(self):
    for k, v in self.contexts.items():
      if 'name' not in v:
        v['name'] = k
    self.contexts = {k: Context(**v) for k, v in self.contexts.items()}

  @classmethod
  def from_yaml(cls, filename: str = DEFAULT_CONFIG):
    with open(filename, 'r') as f:
      cfg = yaml.safe_load(f)
    return cls(**cfg, filename=filename)

  def to_dict(self):
    return {
        'current_context': self.current_context,
        'contexts': {k: v.to_serializable_dict()
                     for k, v in self.contexts.items()}
    }

  def to_yaml(self, filename: str = None):
    if filename is None:
      filename = self.filename
    dir = os.path.dirname(filename)
    if len(dir):
      os.makedirs(dir, exist_ok=True)
    _dict = self.to_dict()
    for k, v in _dict['contexts'].items():
      v.pop('name', None)
    with open(filename, 'w') as f:
      yaml.safe_dump(_dict, f)

  @property
  def current(self) -> Context:
    """ get the current Context """
    return self.contexts[self.current_context]
